@@ -25,6 +25,8 @@
 
 #include "defs.h"
 
+#define USE_SAFE_CHDIR 1
+
 #include <errno.h>
 #include <assert.h>
 
@@ -67,9 +69,9 @@ static void init_mounted_dev_list(void);
 static void process_top_path PARAMS((char *pathname));
 static int process_path PARAMS((char *pathname, char *name, boolean leaf, char *parent));
 static void process_dir PARAMS((char *pathname, char *name, int pathlen, struct stat *statp, char *parent));
-#if 0
-static boolean no_side_effects PARAMS((struct predicate *pred));
-#endif
+
+
+
 static boolean default_prints PARAMS((struct predicate *pred));
 
 /* Name this program was run with. */
@@ -186,6 +188,30 @@ following_links(void)
     }
 }
 
+
+static int
+fallback_stat(const char *name, struct stat *p, int prev_rv)
+{
+  /* Our original stat() call failed.  Perhaps we can't follow a
+   * symbolic link.  If that might be the problem, lstat() the link. 
+   * Otherwise, admit defeat. 
+   */
+  switch (errno)
+    {
+    case ENOENT:
+    case ENOTDIR:
+      return lstat(name, p);
+
+    case EACCES:
+    case EIO:
+    case ELOOP:
+    case ENAMETOOLONG:
+    case EOVERFLOW:
+    default:
+      return prev_rv;	       
+    }
+}
+
 
 /* optionh_stat() implements the stat operation when the -H option is
  * in effect.
@@ -205,16 +231,11 @@ optionh_stat(const char *name, struct stat *p)
       /* This file is from the command line; deference the link (if it
        * is a link).  
        */
-      if (0 == stat(name, p))
-	{
-	  /* success */
-	  return 0;
-	}
+      int rv = stat(name, p);
+      if (0 == rv)
+	return 0;		/* success */
       else
-	{
-	  /* fallback - return the information for the link itself. */
-	  return lstat(name, p);
-	}
+	return fallback_stat(name, p, rv);
     }
   else
     {
@@ -231,14 +252,11 @@ optionh_stat(const char *name, struct stat *p)
 int 
 optionl_stat(const char *name, struct stat *p)
 {
-  if (0 == stat(name, p))
-    {
-      return 0;			/* normal case. */
-    }
+  int rv = stat(name, p);
+  if (0 == rv)
+    return 0;			/* normal case. */
   else
-    {
-      return lstat(name, p);	/* can't follow link, return the link itself. */
-    }
+    return fallback_stat(name, p, rv);
 }
 
 /* optionp_stat() implements the stat operation when the -P option is
@@ -275,12 +293,12 @@ set_follow_state(enum SymlinkOption opt)
     {
     case SYMLINK_ALWAYS_DEREF:  /* -L */
       xstat = optionl_stat;
-      no_leaf_check = false;
+      no_leaf_check = true;
       break;
       
     case SYMLINK_NEVER_DEREF:	/* -P (default) */
       xstat = optionp_stat;
-      /* Can't turn on no_leaf_check because the user might have specified 
+      /* Can't turn no_leaf_check off because the user might have specified 
        * -noleaf anyway
        */
       break;
@@ -289,6 +307,8 @@ set_follow_state(enum SymlinkOption opt)
       xstat = optionh_stat;
       no_leaf_check = true;
     }
+
+  symlink_handling = opt;
   
   /* For DBEUG_STAT, the choice is made at runtime within debug_stat()
    * by checking the contents of the symlink_handling variable.
@@ -332,7 +352,6 @@ main (int argc, char **argv)
   start_time = time (NULL);
   cur_day_start = start_time - DAYSECS;
   full_days = false;
-  no_leaf_check = false;
   stay_on_filesystem = false;
   ignore_readdir_race = false;
   exit_status = 0;
@@ -354,6 +373,7 @@ main (int argc, char **argv)
       error (1, 0, _("The environment variable FIND_BLOCK_SIZE is not supported, the only thing that affects the block size is the POSIXLY_CORRECT environment variable"));
     }
 
+  no_leaf_check = false;
   set_follow_state(SYMLINK_NEVER_DEREF); /* The default is equivalent to -P. */
 
   init_mounted_dev_list();
@@ -551,93 +571,14 @@ specific_dirname(const char *dir)
     }
   else
     {
-      return canonicalize_filename_mode(dir, CAN_EXISTING);
+      char *result = canonicalize_filename_mode(dir, CAN_EXISTING);
+      if (NULL == result)
+	return strdup(dir);
+      else
+	return result;
     }
 }
 
-
-
-
-#if 0
-/* list_item_present: Search for NEEDLE in HAYSTACK.
- *
- * NEEDLE is a normal C string.  HAYSTACK is a list of concatenated C strings, 
- * each with a terminating NUL.   The last item in the list is identified by 
- * the fact that its terminating NUL is itself followed by a second NUL.
- *
- * This data structure does not lend itself to fast searching, but we only
- * do this when wd_sanity_check() thinks that a filesystem might have been 
- * mounted or unmounted.   That doesn't happen very often.
- */
-static int
-list_item_present(const char *needle, const char *haystack)
-{
-  if (NULL != haystack)
-    {
-      const char *s = haystack;
-      while (*s)
-	{
-	  if (0 == strcmp(s, needle))
-	    {
-	      return 1;
-	    }
-	  else
-	    {
-	      s += strlen(s);
-	      ++s;			/* skip the first NUL. */
-	    }
-	}
-    }
-  return 0;
-}
-
-static char *mount_points = NULL;
-
-
-/* Initialise our idea of what the list of mount points is. 
- * this function is called exactly once.
- */
-static void
-init_mount_point_list(void)
-{
-  assert(NULL == mount_points);
-  mount_points = get_mounted_filesystems();
-}
-
-static void
-refresh_mount_point_list(void)
-{
-  if (mount_points)
-    {
-      free(mount_points);
-      mount_points = NULL;
-    }
-  init_mount_point_list();
-}
-
-/* Determine if a directory has recently had a filesystem 
- * mounted on it or unmounted from it.
- */
-static enum MountPointStateChange
-get_mount_point_state(const char *dir)
-{
-  int was_mounted, is_mounted;
-
-  was_mounted = list_item_present(dir, mount_points);
-  refresh_mount_point_list();
-  is_mounted = list_item_present(dir, mount_points);
-
-  if (was_mounted == is_mounted)
-    return MountPointStateUnchanged;
-  else if (is_mounted)
-    return MountPointRecentlyMounted;
-  else 
-    return MountPointRecentlyUnmounted;
-}
-#endif
-
-
-
 static dev_t *mounted_devices = NULL;
 static size_t num_mounted_devices = 0u;
 
@@ -704,6 +645,21 @@ get_mount_state(dev_t newdev)
     return MountPointRecentlyUnmounted;
 }
 
+
+/* Return non-zero if FS is the name of a filesystem that is likely to
+ * be automounted
+ */
+static int
+fs_likely_to_be_automounted(const char *fs)
+{
+  return ( (0==strcmp(fs, "nfs")) || (0==strcmp(fs, "autofs")));
+}
+
+enum WdSanityCheckFatality
+  {
+    FATAL_IF_SANITY_CHECK_FAILS,
+    NON_FATAL_IF_SANITY_CHECK_FAILS
+  };
 
 
 /* Examine the results of the stat() of a directory from before we
@@ -734,30 +690,34 @@ get_mount_state(dev_t newdev)
  * Hence we print a warning message to indicate that the output of find 
  * might be inconsistent due to the change in the filesystem.
  */
-static void
+static boolean
 wd_sanity_check(const char *thing_to_stat,
 		const char *program_name,
 		const char *what,
-		const struct stat *oldinfo,
+		dev_t old_dev,
+		ino_t old_ino,
 		struct stat *newinfo,
 		int parent,
 		int line_no,
 		enum TraversalDirection direction,
+		enum WdSanityCheckFatality isfatal,
 		boolean *changed) /* output parameter */
 {
   const char *fstype;
   char *specific_what = NULL;
-  int isfatal = 1;
+  int silent = 0;
   
   *changed = false;
   
   if ((*xstat) (".", newinfo) != 0)
     error (1, errno, "%s", thing_to_stat);
   
-  if (oldinfo->st_dev != newinfo->st_dev)
+  if (old_dev != newinfo->st_dev)
     {
       *changed = true;
       specific_what = specific_dirname(what);
+      fstype = filesystem_type(thing_to_stat, ".", newinfo);
+      silent = fs_likely_to_be_automounted(fstype);
       
       /* This condition is rare, so once we are here it is 
        * reasonable to perform an expensive computation to 
@@ -784,37 +744,45 @@ wd_sanity_check(const char *thing_to_stat,
 	  switch (transition)
 	    {
 	    case MountPointRecentlyUnmounted:
-	      isfatal = 0;
-	      error (0, 0,
-		     _("Warning: filesystem %s has recently been unmounted."),
-		     specific_what);
+	      isfatal = NON_FATAL_IF_SANITY_CHECK_FAILS;
+	      if (!silent)
+		{
+		  error (0, 0,
+			 _("Warning: filesystem %s has recently been unmounted."),
+			 specific_what);
+		}
 	      break;
 	      
 	    case MountPointRecentlyMounted:
-	      isfatal = 0;
-	      error (0, 0,
-		     _("Warning: filesystem %s has recently been mounted."),
-		     specific_what);
+	      isfatal = NON_FATAL_IF_SANITY_CHECK_FAILS;
+	      if (!silent)
+		{
+		  error (0, 0,
+			 _("Warning: filesystem %s has recently been mounted."),
+			 specific_what);
+		}
 	      break;
 
 	    case MountPointStateUnchanged:
-	      isfatal = 1;
+	      /* leave isfatal as it is */
 	      break;
 	    }
 	}
 
-      if (isfatal)
+      if (FATAL_IF_SANITY_CHECK_FAILS == isfatal)
 	{
 	  fstype = filesystem_type(thing_to_stat, ".", newinfo);
-	  error (isfatal, 0,
+	  error (1, 0,
 		 _("%s%s changed during execution of %s (old device number %ld, new device number %ld, filesystem type is %s) [ref %ld]"),
 		 specific_what,
 		 parent ? "/.." : "",
 		 program_name,
-		 (long) oldinfo->st_dev,
+		 (long) old_dev,
 		 (long) newinfo->st_dev,
 		 fstype,
 		 line_no);
+	  /*NOTREACHED*/
+	  return false;
 	}
       else
 	{
@@ -824,79 +792,172 @@ wd_sanity_check(const char *thing_to_stat,
 	   * without checking the inode number.
 	   */
 	  free(specific_what);
-	  return;
+	  return true;
 	}
     }
 
   /* Device number was the same, check if the inode has changed. */
-  if (oldinfo->st_ino != newinfo->st_ino)
+  if (old_ino != newinfo->st_ino)
     {
       *changed = true;
       specific_what = specific_dirname(what);
       fstype = filesystem_type(thing_to_stat, ".", newinfo);
       
-      error (1, 0,
+      error ((isfatal == FATAL_IF_SANITY_CHECK_FAILS) ? 1 : 0,
+	     0,			/* no relevant errno value */
 	     _("%s%s changed during execution of %s (old inode number %ld, new inode number %ld, filesystem type is %s) [ref %ld]"),
 	     specific_what, 
 	     parent ? "/.." : "",
 	     program_name,
-	     (long) oldinfo->st_ino,
+	     (long) old_ino,
 	     (long) newinfo->st_ino,
 	     fstype,
 	     line_no);
       free(specific_what);
+      return false;
     }
+  
+  return true;
 }
 
 enum SafeChdirStatus
   {
     SafeChdirOK,
     SafeChdirFailSymlink,
+    SafeChdirFailNotDir,
     SafeChdirFailStat,
-    SafeChdirFailDefect
+    SafeChdirFailWouldBeUnableToReturn,
+    SafeChdirFailChdirFailed,
+    SafeChdirFailNonexistent
   };
 
-/* Safely perform a change in directory. */
+/* Safely perform a change in directory.
+ *
+ */
 static int 
-safely_chdir(const char *dest, enum TraversalDirection direction)
+safely_chdir(const char *dest,
+	     enum TraversalDirection direction,
+	     struct stat *statbuf_dest)
 {
-  struct stat statbuf_dest, statbuf_arrived;
-  int rv=SafeChdirFailDefect, dotfd=-1;
+  struct stat statbuf_arrived;
+  int rv, dotfd=-1;
+  int saved_errno;		/* specific_dirname() changes errno. */
   char *name = NULL;
-
-  errno = 0;
+  boolean rv_set = false;
+  
+  saved_errno = errno = 0;
   dotfd = open(".", O_RDONLY);
   if (dotfd >= 0)
     {
       /* Stat the directory we're going to. */
-      if (0 == (following_links() ? stat : lstat)(dest, &statbuf_dest))
+      if (0 == xstat(dest, statbuf_dest))
 	{
 #ifdef S_ISLNK
-	  if (!following_links() && S_ISLNK(statbuf_dest.st_mode))
+	  if (!following_links() && S_ISLNK(statbuf_dest->st_mode))
 	    {
 	      rv = SafeChdirFailSymlink;
+	      rv_set = true;
+	      saved_errno = 0;	/* silence the error message */
 	      goto fail;
 	    }
 #endif	  
-	  chdir(dest);
+#ifdef S_ISDIR
+	  /* Although the immediately following chdir() would detect
+	   * the fact that this is not a directory for us, this would
+	   * result in an extra system call that fails.  Anybody
+	   * examining the system-call trace should ideally not be
+	   * concerned that something is actually failing.
+	   */
+	  if (!S_ISDIR(statbuf_dest->st_mode))
+	    {
+	      rv = SafeChdirFailNotDir;
+	      rv_set = true;
+	      saved_errno = 0;	/* silence the error message */
+	      goto fail;
+	    }
+#endif	  
+	  if (0 == chdir(dest))
+	    {
+	      /* check we ended up where we wanted to go */
+	      boolean changed = false;
+	      wd_sanity_check(".", program_name, ".",
+			      statbuf_dest->st_dev,
+			      statbuf_dest->st_ino,
+			      &statbuf_arrived, 
+			      0, __LINE__, direction,
+			      FATAL_IF_SANITY_CHECK_FAILS,
+			      &changed);
+	      close(dotfd);
+	      return SafeChdirOK;
+	    }
+	  else
+	    {
+	      saved_errno = errno;
+	      if (ENOENT == saved_errno)
+		{
+		  rv = SafeChdirFailNonexistent;
+		  rv_set = true;
+		  if (ignore_readdir_race)
+		    errno = 0;	/* don't issue err msg */
+		  else
+		    name = specific_dirname(dest);
+		}
+	      else if (ENOTDIR == saved_errno)
+		{
+		  /* This can happen if the we stat a directory,
+		   * and then filesystem activity changes it into 
+		   * a non-directory.
+		   */
+		  saved_errno = 0;	/* don't issue err msg */
+		  rv = SafeChdirFailNotDir;
+		  rv_set = true;
+		}
+	      else
+		{
+		  rv = SafeChdirFailChdirFailed;
+		  rv_set = true;
+		  name = specific_dirname(dest);
+		}
+	      goto fail;
+	    }
 	}
       else
 	{
+	  saved_errno = errno;
 	  rv = SafeChdirFailStat;
+	  rv_set = true;
 	  name = specific_dirname(dest);
+	  if ( (ENOENT == saved_errno) || (0 == curdepth))
+	    saved_errno = 0;	/* don't issue err msg */
 	  goto fail;
 	}
     }
   else
     {
+      /* We do not have read permissions on "." */
+      rv = SafeChdirFailWouldBeUnableToReturn;
+      rv_set = true;
+      goto fail;
     }
+
+  saved_errno = 0;
   
+  /* We use the same exit path for successs or failure. 
+   * which has occurred is recorded in RV. 
+   */
  fail:
-  if (errno)
+  if (saved_errno)
     {
+      errno = saved_errno;
+      
+      /* do not call error() as this would result in a duplicate error message 
+       * when the caller does the same thing. 
+       */
+#if 0
       if (NULL == name)
 	name = specific_dirname(".");
-      error(0, errno, "%s", name);
+      error(0, saved_errno, "%s", name);
+#endif
     }
   
   free(name);
@@ -907,6 +968,7 @@ safely_chdir(const char *dest, enum TraversalDirection direction)
       close(dotfd);
       dotfd = -1;
     }
+  assert(rv_set);
   return rv;
 }
 
@@ -924,9 +986,14 @@ chdir_back (void)
 	error (1, errno, "%s", starting_dir);
 
       wd_sanity_check(starting_dir,
-		      program_name, starting_dir,
-		      &starting_stat_buf, &stat_buf, 0, __LINE__,
-		      TraversingUp, &dummy);
+		      program_name,
+		      starting_dir,
+		      starting_stat_buf.st_dev,
+		      starting_stat_buf.st_ino,
+		      &stat_buf, 0, __LINE__,
+		      TraversingUp,
+		      FATAL_IF_SANITY_CHECK_FAILS,
+		      &dummy);
     }
   else
     {
@@ -936,9 +1003,15 @@ chdir_back (void)
 }
 
 /* Descend PATHNAME, which is a command-line argument.  */
-
 static void
 process_top_path (char *pathname)
+{
+  process_path (pathname, pathname, false, ".");
+}
+
+
+static void
+old_process_top_path (char *pathname)
 {
   struct stat stat_buf, cur_stat_buf;
   boolean dummy;
@@ -950,6 +1023,43 @@ process_top_path (char *pathname)
      once here and once in process_path.  It's not too bad, though,
      since the kernel can read the stat information out of its inode
      cache the second time.  */
+#if USE_SAFE_CHDIR
+  if ((*xstat) (pathname, &stat_buf) == 0 && S_ISDIR (stat_buf.st_mode))
+    {
+      enum SafeChdirStatus rv = safely_chdir(pathname, TraversingDown, &stat_buf);
+      
+      switch (rv)
+	{
+	case SafeChdirOK:
+	  process_path (pathname, ".", false, ".");
+	  chdir_back ();
+	  return;
+	  
+	case SafeChdirFailNonexistent:
+	case SafeChdirFailStat:
+	case SafeChdirFailWouldBeUnableToReturn:
+	case SafeChdirFailSymlink:
+	case SafeChdirFailNotDir:
+	case SafeChdirFailChdirFailed:
+	  if ((SafeChdirFailNonexistent==rv) && !ignore_readdir_race)
+	    {
+	      error (0, errno, "%s", pathname);
+	      exit_status = 1;
+	    }
+	  else
+	    {
+	      process_path (pathname, pathname, false, ".");
+	    }
+	  chdir_back ();
+	  return;
+	}
+    }
+  else
+    {
+      /* Not a directory */
+      process_path (pathname, pathname, false, ".");
+    }
+#else
   if ((*xstat) (pathname, &stat_buf) == 0 && S_ISDIR (stat_buf.st_mode))
     {
       if (chdir (pathname) < 0)
@@ -965,14 +1075,21 @@ process_top_path (char *pathname)
       /* Check that we are where we should be. */
       wd_sanity_check(pathname, program_name,
 		      ".",
-		      &stat_buf, &cur_stat_buf, 0, __LINE__,
-		      TraversingDown, &dummy);
+		      stat_buf.st_dev,
+		      stat_buf.st_ino,
+		      &cur_stat_buf, 0, __LINE__,
+		      TraversingDown,
+		      FATAL_IF_SANITY_CHECK_FAILS,
+		      &dummy);
 
       process_path (pathname, ".", false, ".");
       chdir_back ();
     }
   else
-    process_path (pathname, pathname, false, ".");
+    {
+      process_path (pathname, pathname, false, ".");
+    }
+#endif
 }
 
 /* Info on each directory in the current tree branch, to avoid
@@ -991,6 +1108,53 @@ static int dir_curr = -1;
 /* (Arbitrary) number of entries to grow `dir_ids' by.  */
 #define DIR_ALLOC_STEP 32
 
+
+
+/* We've detected a filesystem loop.   This is caused by one of 
+ * two things:
+ *
+ * 1. Option -L is in effect and we've hit a symbolic link that 
+ *    points to an ancestor.  This is harmless.  We won't traverse the 
+ *    symbolic link.
+ *
+ * 2. We have hit a real cycle in the directory hierarchy.  In this 
+ *    case, we issue a diagnostic message (POSIX requires this) and we
+ *    skip that directory entry.
+ */
+static void
+issue_loop_warning(const char *name, const char *pathname, int level)
+{
+  struct stat stbuf_link;
+  if (lstat(name, &stbuf_link) != 0)
+    stbuf_link.st_mode = S_IFREG;
+  
+  if (S_ISLNK(stbuf_link.st_mode))
+    {
+      error(0, 0,
+	    _("Symbolic link `%s' is part of a loop in the directory hierarchy; we have already visited the directory to which it points."),
+	    pathname);
+    }
+  else
+    {
+      int distance = 1 + (dir_curr-level);
+      /* We have found an infinite loop.  POSIX requires us to
+       * issue a diagnostic.  Usually we won't get to here
+       * because when the leaf optimisation is on, it will cause
+       * the subdirectory to be skipped.  If /a/b/c/d is a hard
+       * link to /a/b, then the link count of /a/b/c is 2,
+       * because the ".." entry of /b/b/c/d points to /a, not
+       * to /a/b/c.
+       */
+      error(0, 0,
+	    _("Filesystem loop detected; `%s' has the same device number and inode as a directory which is %d %s."),
+	    pathname,
+	    distance,
+	    (distance == 1 ?
+	     _("level higher in the filesystem hierarchy") :
+	     _("levels higher in the filesystem hierarchy")));
+    }
+}
+
 /* Recursively descend path PATHNAME, applying the predicates.
    LEAF is true if PATHNAME is known to be in a directory that has no
    more unexamined subdirectories, and therefore it is not a directory.
@@ -1049,8 +1213,11 @@ process_path (char *pathname, char *name, boolean leaf, char *parent)
   for (i = 0; i <= dir_curr; i++)
     if (stat_buf.st_ino == dir_ids[i].ino &&
 	stat_buf.st_dev == dir_ids[i].dev)
-      stop_at_current_level = true;
-
+      {
+	stop_at_current_level = true;
+	issue_loop_warning(name, pathname, i);
+      }
+  
   if (dir_alloc <= ++dir_curr)
     {
       dir_alloc += DIR_ALLOC_STEP;
@@ -1115,13 +1282,9 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
   name_space = savedir (name);
   if (name_space == NULL)
     {
-      if (errno)
-	{
-	  error (0, errno, "%s", pathname);
-	  exit_status = 1;
-	}
-      else
-	error (1, 0, _("virtual memory exhausted"));
+      assert(errno != 0);
+      error (0, errno, "%s", pathname);
+      exit_status = 1;
     }
   else
     {
@@ -1139,6 +1302,35 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
       cur_path_size = 0;
       cur_path = NULL;
 
+#if USE_SAFE_CHDIR
+      if (strcmp (name, "."))
+	{
+	  enum SafeChdirStatus status = safely_chdir (name, TraversingDown, &stat_buf);
+	  switch (status)
+	    {
+	    case SafeChdirOK:
+	      /* If there had been a change but wd_sanity_check()
+	       * accepted it, we need to accept that on the 
+	       * way back up as well, so modify our record 
+	       * of what we think we should see later.
+	       * If there was no change, the assignments are a no-op.
+	       */
+	      dir_ids[dir_curr].dev = stat_buf.st_dev;
+	      dir_ids[dir_curr].ino = stat_buf.st_ino;
+	      break;
+      
+	    case SafeChdirFailNonexistent:
+	    case SafeChdirFailStat:
+	    case SafeChdirFailWouldBeUnableToReturn:
+	    case SafeChdirFailSymlink:
+	    case SafeChdirFailNotDir:
+	    case SafeChdirFailChdirFailed:
+	      error (0, errno, "%s", pathname);
+	      exit_status = 1;
+	      return;
+	    }
+	}
+#else
       if (strcmp (name, ".") && chdir (name) < 0)
 	{
 	  error (0, errno, "%s", pathname);
@@ -1149,18 +1341,16 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
       /* Check that we are where we should be. */
       if (1)
 	{
-	  struct stat tmp;
-	  boolean changed;
-	  
-	  memset(&tmp, 0, sizeof(tmp));
-	  tmp.st_dev = dir_ids[dir_curr].dev;
-	  tmp.st_ino = dir_ids[dir_curr].ino;
-	  changed = false;
+	  boolean changed = false;
 	  wd_sanity_check(pathname,
 			  program_name,
 			  ".",
-			  &tmp, &stat_buf, 0, __LINE__, 
-			  TraversingDown, &changed);
+			  dir_ids[dir_curr].dev,
+			  dir_ids[dir_curr].ino,
+			  &stat_buf, 0, __LINE__, 
+			  TraversingDown,
+			  FATAL_IF_SANITY_CHECK_FAILS,
+			  &changed);
 	  if (changed)
 	    {
 	      /* If there had been a change but wd_sanity_check()
@@ -1172,6 +1362,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	      dir_ids[dir_curr].ino = stat_buf.st_ino;
 	    }
 	}
+#endif
 
       for (namep = name_space; *namep; namep += file_len - pathname_len + 1)
 	{
@@ -1208,7 +1399,66 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	    process_path (cur_path, cur_name, false, pathname);
 	  curdepth--;
 	}
+#if USE_SAFE_CHDIR
+      if (strcmp (name, "."))
+	{
+	  enum SafeChdirStatus status;
+	  struct dir_id did;
+	  boolean changed = false;
+	  
+	  /* We could go back and do the next command-line arg
+	     instead, maybe using longjmp.  */
+	  char const *dir;
+	  boolean deref = following_links() ? true : false;
+	  
+	  if ( (curdepth>0) && !deref)
+	    dir = "..";
+	  else
+	    {
+	      chdir_back ();
+	      dir = parent;
+	    }
+	  
+	  status = safely_chdir (dir, TraversingUp, &stat_buf);
+	  switch (status)
+	    {
+	    case SafeChdirOK:
+	      break;
+      
+	    case SafeChdirFailNonexistent:
+	    case SafeChdirFailStat:
+	    case SafeChdirFailWouldBeUnableToReturn:
+	    case SafeChdirFailSymlink:
+	    case SafeChdirFailNotDir:
+	    case SafeChdirFailChdirFailed:
+	      error (1, errno, "%s", pathname);
+	      return;
+	    }
 
+	  if (dir_curr > 0)
+	    {
+	      did.dev = dir_ids[dir_curr-1].dev;
+	      did.ino = dir_ids[dir_curr-1].ino;
+	    }
+	  else
+	    {
+	      did.dev = starting_stat_buf.st_dev;
+	      did.ino = starting_stat_buf.st_ino;
+	    }
+
+	  wd_sanity_check(pathname,
+			  program_name,
+			  parent,
+			  did.dev,
+			  did.ino,
+			  &stat_buf,
+			  deref ? 1 : 0,
+			  __LINE__, 
+			  TraversingUp,
+			  FATAL_IF_SANITY_CHECK_FAILS,
+			  &changed);
+	}
+#else
       if (strcmp (name, "."))
 	{
 	  /* We could go back and do the next command-line arg
@@ -1250,11 +1500,16 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	      wd_sanity_check(pathname,
 			      program_name,
 			      parent,
-			      &tmp, &stat_buf,
+			      tmp.st_dev,
+			      tmp.st_ino,
+			      &stat_buf,
 			      problem_is_with_parent, __LINE__, 
-			      TraversingUp, &changed);
+			      TraversingUp,
+			      FATAL_IF_SANITY_CHECK_FAILS,
+			      &changed);
 	    }
 	}
+#endif
 
       if (cur_path)
 	free (cur_path);
@@ -1262,23 +1517,6 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
     }
 }
 
-#if 0
-/* Return true if there are no side effects in any of the predicates in
-   predicate list PRED, false if there are any. */
-
-static boolean
-no_side_effects (struct predicate *pred)
-{
-  while (pred != NULL)
-    {
-      if (pred->side_effects)
-	return (false);
-      pred = pred->pred_next;
-    }
-  return (true);
-}
-#endif
-
 /* Return true if there are no predicates with no_default_print in
    predicate list PRED, false if there are any.
    Returns true if default print should be performed */
