@@ -193,8 +193,10 @@ static size_t rplen = 0;
 
 /* If nonzero, when this string is read on stdin it is treated as
    end of file.
-   I don't like this - it should default to NULL.  */
-static char *eof_str = "_";
+   IEEE Std 1003.1, 2004 Edition allows this to be NULL.
+   In findutils releases up to and including 4.2.8, this was "_".
+*/
+static char *eof_str = NULL;
 
 /* If nonzero, the maximum number of nonblank lines from stdin to use
    per command line.  */
@@ -265,7 +267,7 @@ static struct option const longopts[] =
   {"null", no_argument, NULL, '0'},
   {"arg-file", required_argument, NULL, 'a'},
   {"eof", optional_argument, NULL, 'e'},
-  {"replace", optional_argument, NULL, 'i'},
+  {"replace", optional_argument, NULL, 'I'},
   {"max-lines", optional_argument, NULL, 'l'},
   {"max-args", required_argument, NULL, 'n'},
   {"interactive", no_argument, NULL, 'p'},
@@ -410,7 +412,7 @@ main (int argc, char **argv)
   
 
   
-  while ((optc = getopt_long (argc, argv, "+0a:e::i::l::n:prs:txP:",
+  while ((optc = getopt_long (argc, argv, "+0a:E::e::i::l::n:prs:txP:",
 			      longopts, (int *) 0)) != -1)
     {
       switch (optc)
@@ -419,8 +421,9 @@ main (int argc, char **argv)
 	  read_args = read_string;
 	  break;
 
-	case 'e':
-	  if (optarg)
+	case 'E':		/* POSIX */
+	case 'e':		/* deprecated */
+	  if (optarg && (strlen(optarg) > 0))
 	    eof_str = optarg;
 	  else
 	    eof_str = 0;
@@ -430,7 +433,8 @@ main (int argc, char **argv)
 	  usage (stdout);
 	  return 0;
 
-	case 'i':
+	case 'I':		/* POSIX */
+	case 'i':		/* deprecated */
 	  if (optarg)
 	    replace_pat = optarg;
 	  else
@@ -440,7 +444,8 @@ main (int argc, char **argv)
 	  lines_per_exec = 0;
 	  break;
 
-	case 'l':
+	case 'L':		/* POSIX */
+	case 'l':		/* deprecated */
 	  if (optarg)
 	    lines_per_exec = parse_num (optarg, 'l', 1L, -1L, 1);
 	  else
@@ -989,6 +994,16 @@ push_arg (char *arg, size_t len)
       cmd_argv[cmd_argc++] = argbuf + cmd_argv_chars;
       strcpy (argbuf + cmd_argv_chars, arg);
       cmd_argv_chars += len;
+      
+      /* If we have now collected enough arguments,
+       * do the exec immediately.  This must be 
+       * conditional on arg!=NULL, since do_exec() 
+       * actually calls push_arg(NULL, 0).
+       */
+      if (!initial_args
+	  && args_per_exec
+	  && (cmd_argc - initial_argc) == args_per_exec)
+	do_exec ();
     }
 }
 
